
/*		CFSCTYPES.H for the CFS-C Classifier System.

Definitions of data structures for the "core" & "dsclearn"  CFS-C classifier system components:
	the classifiers and classifier list
	the messages and the message lists (current and new)
	the 'candidate' classifier structures and list -- these are just classifier nodes now.
	the 'candidate' matches - triples of pointers to (classifier, message-1, message-2).
	the effector interface (at least the domain independent part thereof).
	And all Variables needed by the discovery algorithms.

**/

#ifndef _CFSCTYPES_H
#define _CFSCTYPES_H 1
/* ------------------------------- */

#include "config.h"
#include "core.h"
#include "dsclearn.h"

/*  MsgNode - structure to store a message and associated attributes:  */

struct MsgNode {
    unsigned int Msg_Id;		/* Id number for the message. */
    unsigned int Message[INTPRSTRMAX];	/* The message */
    float Intensit;			/* Bid payed to post this message. */
    struct CfNode *Producer;		/* message's producer. */

    struct MsgNode *MtchMsg1;		/* message matching condition-1 of 'Matched' classifier */
    struct MsgNode *MtchMsg2;		/* message matching condition-2 of 'Matched' classifier */
    struct MsgNode *NxtEMsg;		/* next message matching an effector. */

    short FromDet;			/* Set TRUE if message came from detectors. */
    short IsMsg;			/* Set FALSE if this node is a deleted message (by an effector). */
    short MsgMtchd;			/* Set TRUE if message matches some classifier condition */
    short MsgUsed;			/* Set TRUE if message is used by a classifier. */
};


/* CfNode  -- structure to store a classifier and associated attributes: */

struct CfNode {
    unsigned int Cf_Id;

   /*   Condition  1   and 2  */

    unsigned int Cnd1Bits[INTPRSTRMAX];	/* 0/1 bits -> 0,1  # -> 0 */
    unsigned int Cnd1DCs[INTPRSTRMAX];	/* DC #'s   -> 0,   else 1 */
    short Cnd1Type;			/* 0 is match, 1 is not-match */
    unsigned int Cnd2Bits[INTPRSTRMAX];	/* As for condition 1 */
    unsigned int Cnd2DCs[INTPRSTRMAX];
    short Cnd2Type;

   /* The Action part */

    short ActType;			/* 0 for stardard pass-through (overlay). */
    unsigned int ActBits[INTPRSTRMAX];	/* 0,1 bits -> 0,1  # -> 0 */
    unsigned int ActDCs[INTPRSTRMAX];	/* DC #'s -> 1,  else 0 */

   /* Longer-lived variable values */

    float Strength;			/* Accumulated strength of classifier */
    float BidRatio;			/* Number of dont-cares (#'s) */
    unsigned int TotNmBid;		/* Total number of times it bid. */
    unsigned int TotMtch;		/* Total number of 'matches' (candidate msgs) */
    unsigned int TotProd;		/* Total number of messages produced. */
    unsigned int TotPost;		/* Total number of messages posted (AFTER effector conflist resolution.) */
    unsigned int TotEMtch;		/* Total number of messages produced that matched effectors. */
    unsigned int TotEAct;		/* Total number of messages produced that activated effectors. */
    unsigned int TotPosRw;		/* Total number of times activate when SysRew > 0 */
    unsigned int TotNegRw;		/* Total number of times activate when SysRew < 0 */
    unsigned int TotNmOfs;		/* Total number of offspring */

    /* Variables used by the PSP: */
    float PSPBid;	/* Bid made by Cf when last active */
    unsigned int PSPId; /* Id is stored here to prevent us from paying to Cfs that no longer exist. */

   /* Temporary variables, for use in bidding by candidates and by bucket brigade. */
    float ChngStr;			/* Change in strength at last step */
    float Support;			/* Sum of Intensities of match msgs */
    float CfBid;			/* bid made by classifier. */
    float CfBBid;			/* 'Biased' bid - used to calc. probability of winning */
    float BidShare;			/* Share of bid to pay suppliers of matched messages */
    short PstEfMsg;			/* Number of effector messages POSTED by this classifier */

    unsigned int NmMtch;		/* Number of candidate-matches this cf is in */
    unsigned int MMBitsC1[INTPRML];	/* Store 1 bit / message - set bit to 1 when msg matches conditon 1 */
    short NmMMC1;			/* Number of messages that match condition 1 */
    unsigned int MMBitsC2[INTPRML];	/* Store 1 bit / message - set bit to 1 when msg matches conditon 2 */
    short NmMMC2;			/* Number of messages that match condition 2 */

    short NmProd;			/* Num. of msgs produced (number BEFORE effector conflict resolution. */
    short NmPost;			/* Num. of msgs AFTER effector conflict resolution--posted to step t+1. */
    short NmDetMM;			/* Num of Detector Messages used to post NmPost output messages:
											used to calculate bid share */

   /* Variables for use by 'Discovery' algorithms */

    float PBiasStr;			/* Biased strength for prob. being a Parent */
    float RBiasStr;			/* Biased strength for prob. being a Replaced */
    short EligRpl;			/* TRUE => This is eligible to be replaced this step */
    short ReplFlg;			/* TRUE => This chosen to be in replace-able 'pool' this step */
    unsigned int StpCrtd;		/* Step classifier created */
    unsigned int StpLPrd;		/* Step cf last 'produced' msgs (won the bidding competition) */
    unsigned int StpLPst;		/* Step cf last 'posted' msgs (after effector conflict resolution) */

    short NoBBFlg;			/* TRUE => don't apply BB to this classifier. */
    short NoRplFlg;			/* TRUE => don't replace this classifier. */
    short NoPrnFlg;			/* TRUE => don't use as parent */

   /* Pointers for linked lists running through the CfNodes */

    struct CfNode *NxtCand;		/* next candidate classifier. */

    struct CfNode *PrvStrCf;		/* For 2-way linked list of classifiers, sorted on strength */
    struct CfNode *NxtStrCf;		/* from lower strength (Prev) to higher (Next). StrSrtCf is start. */
};


/*  Effectornode: store features associated with an effector: */

struct EffNode {
    unsigned int Eff_Id;
    char EffName[EFFNAMSZ];		/* The name of the effector */

    unsigned int ECndBit[INTPRSTRMAX];	/* Condition to control effector. */
    unsigned int ECndDCs[INTPRSTRMAX];
    unsigned int ECndType;		/* Should be MATCH */

    int (*Eff_Set) ();			/* Function that processes msgs into effector settings */
    int (*Eff_Fn) ();			/* Function that selects an action and does it, given the settings. */
    int (*EQEfAct) ();			/* Function that checks for messages not consistant with action. */
    unsigned int NmEMtch;		/* Number of new messages that match this effector. */
    unsigned int NmEAct;		/* Number of new messages that match msg that activated effector + 1 */
    unsigned int CEffFlg;		/* TRUE => apply DscCEffs() to it to generate new classifiers. */
    struct MsgNode *EffMtch;		/* Pointer to list of candidate msgnodes that satisfy this effector. */

    unsigned int ETotMtch;		/* Cumulative total number of times effector matched */
    unsigned int ETotAct;		/* Cumulative Total number of times effector produces behavior */
};



/* struct NCfNode: store 'business'-part of new classifiers:  */

struct NCfNode { 
    unsigned int NewCf_Id;
    unsigned int NewCnd1B[INTPRSTRMAX];    /* conditon parts */
    unsigned int NewCnd1D[INTPRSTRMAX];
    short NewCnd1T;
    unsigned int NewCnd2B[INTPRSTRMAX];
    unsigned int NewCnd2D[INTPRSTRMAX];
    short NewCnd2T;
    short NewActT;                  /* action part */
    unsigned int NewActB[INTPRSTRMAX];
    unsigned int NewActD[INTPRSTRMAX];
    float NewStr;                       /* strength and bid ratio */
    float NewBR;
};


/* struct ActCfNd: Previously active classifier list node: */

struct ActCfNd {
    struct CfNode *ActCf;		/* Pointer to active cf */
    float ChngStr;			/* Change in strength = Rewards,taxes,BB exchanges */
};


/* struct cfsc_s:  Holds all necessary variables: */

struct cfsc_s {
    /* =1 means factor specifity into payment, =0 means don't */
    int specPay;
    /* Random seed: */
    int	uRndSd;
    /* CycleStp:	Model time clock: the step to be done next.
		(Incremented at end of major-cycle loop in Classifier() in the CFS-C.C file.)	*/
    unsigned int cycleStp ;

    /* Stolen from usercmd.def,  now here (all who need it must include core.ext) */
    unsigned int demoLev ;          /* Control demonstation-output (off=0) */
    short echoFlg ;              /* Control echoing of user commands. */

    /* Stolen from cfsio.def, now here (all who need it must include core.ext) */
    short logFlg;				/* Set ON (1) tells Write macro (see CFSIO.H) to echo to LogFName. */
    short genFlg;				/* Set 1 to have system write 'geneolgy' information. */

    /* Stolen from cfsc.c: */
    /*  Variables to count cycles and decide when to auto-print variables or reports.

    AutoDspl = number of AD...Int items that are > 0 (0 is 'no-display').
                         So if AutoDspl is 0, there is nothing to autodisplay.

    ADxxxInt = display interval for item 'xxx'.  0 is taken as 'no-display'.
    ADxxxFmt = display format     for item 'xxx'.                                */

    short clampFlg ;             /* Set TRUE to stop all learning, AND CycleStp */
    short stopFlag ;             /* Set TRUE to stop run. */

    unsigned int autoDspl ;              /* OFF if ALL autodisplay is off, ON if any one or more are ON */
    unsigned int aDCMInt ;           /* CM = current messages */
    unsigned int aDCMFmt ;
    unsigned int aDDMInt ;           /* DM = detector messages */
    unsigned int aDDMFmt ;
    unsigned int aDCfInt ;           /* Cf = classifiers */
    unsigned int aDCfFmt ;
    unsigned int aDVarInt ;          /* Var = system variables */
    unsigned int aDVarFmt ;
    unsigned int aDCycInt ;          /* Cyc = Cycle Step */
    unsigned int aDCycFmt ;
    unsigned int aDEnvInt ;          /* Env = environment  */
    unsigned int aDEnvFmt ;

    unsigned int saveInt ;           /* AutoSave interval */
    unsigned int saveFNum ;              /* Number of NEXT AutoSave file */

    /* Variables associated with Messages: */
    struct MsgNode *detMsgs[DMSGSSZ];			/* Pointers to messages on CurMsgs (and then OldMsgs) from detectors */

    struct MsgNode *curMsgs,*nextCMsg;  /* Start of current message list. next unused node. */
    unsigned int nmCMsgs ;		/* Number of messages on current message list. */
    unsigned int nmCMsgIn ;		/* Num. of internally genenerated messages. */
    unsigned int nmDetMsg ;		/* Num. of messages from detectors. */
    int lastCMsg ;			/* 'Array' subscript of last msg - allow -1 for empty list */

    struct MsgNode *newMsgs,*nextNMsg;	/* Start of new message list. next unused node. */
    unsigned int nmNMsgs ;		/* Number of messages on new message list. */
    unsigned int nmNMsgIn ;		/* Num. of internally genenerated messages. */
    int lastNMsg ;			/* 'Array' subscript of last msg - allow -1 for empty list */

    struct MsgNode *oldMsgs,*nextOMsg;		/* Start of old message list. to next unused node. */
    unsigned int nmOMsgs ;		/* Number of messages on old message list. */
    unsigned int nmOMsgIn ;		/* Num. of internally genenerated messages. */
    int lastOMsg ;			/* 'Array' subscript of last msg - allow -1 for empty list */

    unsigned int nmIMsgMx ;		/* max. number of 'internally' generated (non-detector) messages. */
    unsigned int nmDMsgMx ;		/* max. number of detector messages allowed at one step. */
    float msgIntDf ;			/* Default message intensity. */

    unsigned int tOTMtch ;		/* Total number of 'matches' */
    unsigned int tOTMsPrd ;		/* Total number of new messages produced (before effector conflict resolution) */
    unsigned int tOTMsPst ;		/* Total number of messages posted (after effector conflict resolution) */
    unsigned int tOTDtMsg ;		/* Total number of messages from detectors */

    /*  Variables associated with classifiers: */
    struct CfNode *cfLst ;		/* Storage for classifier list. */
    struct CfNode *curCfs ;		/* list of current Classifiers. */
    struct CfNode *nxtFrCf ;		/* next free node */
    struct CfNode *candCfs ;		/* list of candidate classifiers. */
    struct CfNode *lstCndCf ;		/* last on CandCfs list. */
    struct CfNode *strSrtCf ;		/* start (low strength) of 2-way list of CfNodes sorted on strength */
    struct CfNode *hiStrCf ;		/* end (high str) of 2-way list of CfNodes sorted on strength */
    struct CfNode *highestStrCf ;	/* Highest strength cf after call to FindHighLowStrCf() in CFSUTIL.C */
    struct CfNode *lowestStrCf ;	/* Lowest strength cf after call ... */

    unsigned int oneMPerC ;		/* 1 => only 1 Msg produced per Classifier */
    unsigned int delHall ;		/* 1 => delete hallucinations (00... msgs). */
    unsigned int mxDupMsg ;		/* Max number of duplicate messages allowed (0 => no max) */

    unsigned int nxtCfId ;		/* Next Cf_Id to use */
    unsigned int nmCfs ;			/* Number of classifiers in currently in list. */
    unsigned int nmCfsMx ;		/* Settable maximum number of classifiers */
    unsigned int nmCandCf ;		/* Number of candidate-producer classifiers. */
    unsigned int nmCfWon ;		/* Num. classifiers that won in bidding (produced messages). */
    unsigned int nmCfPost ;		/* Num. classifiers that posted msgs (AFTER effector conflict resolution) */

    float totCfStr ;			/* Total strength for all current classifiers. */
    float aveCfStr ;			/* Average strength for all current classifiers */
    float totCfBR ;			/* Total bidratio, to calculate average. */
    float aveCfBR ;			/* Average bidration for all current classifiers */

    float cfStrDf ;			/* default classifier strength. */
    float cfSuppDf ;			/* default support increment for NOTMATCH condition 2 */
    float cfStrMax ;			/* maximum allowed strength */
    float cfStrMin ;			/* minimum allowed strength */
    float cfBidMin ;			/* minimum bid allowed */

    /*  Variables for generating classifiers by the GnRndCfs() function */
    float gRCFrMC ;			/* Fraction of 'match' conditions generated by GnRndCfs() function */
    float gRCFrNSL ;			/* Fraction of non-specific loci (DON'T CARES) generated by GnRndCfs() function */

    /* Grand total values for a whole run (for all classifiers, since t=0) */
    float tOTBidRa ;			/* Total bidratio */
    unsigned int tOTNmBid ;		/* Total number of bids */
    unsigned int tOTNmWin ;		/* Total number of winning cfs */
    unsigned int tOTCfPst ;		/* Total number of posting classifiers  (AFTER effector conflict resolution) */
    unsigned int tOTBGStr ;		/* Total number of times bid > strength for cfs with str > 0 */
    unsigned int tOTActEf ;		/* Total numberp of times effectors activated */
    unsigned int tOTNmPRw ;		/* Total number of cfs that got SysRew's > 0 */
    unsigned int tOTNmNRw ;		/* Total number of cfs that got SysRew's < 0 */
    unsigned int tOTSyRwP ;		/* Total number of steps with SysRew > 0 */
    unsigned int tOTSyRwN ;		/* Total number of steps with SysRew < 0 */

    /*   Variables for handling "matches":   */
    unsigned int nmMtchs ;		/* Total number of matches for one cycle. */
    unsigned int nmIEMMx ;		/* Max number of Internally generated Effector Msgs */
    unsigned int nmIEMsgs;			/* number produced during current step */
    unsigned int nmIIMMx ;		/* Max number of Internally generated Internal (non-effector) Msgs */
    unsigned int nmIIMsgs;			/* number produced during current step */

    /* Special variables for bidding and classifier strength modification. */
    unsigned int bBType ;		/* Bucket Brigade Type: 0 = Standard, 1 = Pay all Matches */
    float sysTreas ;		/* System treasury: from taxes, pay for detector msgs, out via SysRew */
    float sysRew ;			/* Global reward to system from Environment. */
    float frPayDet ;			/* Fraction of bid 'paid' for using detector messages */
    float frPayNEC ;			/* Fraction of SysRew 'paid' to non-effector msg producers */
    float dMShare ;			/* Share of Bid to detector messages--the rest to produce classifiers */
    float dMDscnt ;			/* DMDscnt = 1 - DMShare; use to calculate BidShare */
    short shareRew ;			/* Controls how SysRew is distributed. See BBLEARN.C for details. */
    short allPayBd ;			/* 0 = Pay bid only if "post" msg; 1 = Pay if "produce" msg */
    short suppBid ;			/* 1 = Support used to calc. bid, else its not used. */
    float bidPow ;			/* biased-bid is bid raised to this power */
    float bid_k ;			/* bid mulitplier "constant" */
    float bRPow ;			/* Bid Ratio is raised to this power in bid calculation */
    float eBRPow ;			/* BidRatio raised to this power and factored into 'effective' bid */
    float btCSB ;			/* Bias Coupled Stage-setter bid */
    float btCMB ;			/* Bias Coupled memory bid */
    float headTax ;			/* Tax (rate) on all classifiers, once/step */
    float aHeadTax ;			/* Tax (abs amount) on all classifiers, once/step */
    float bidTax ;			/* Tax (rate) on bidding classifiers */
    float fBidTax ;			/* Tax (rate) on bidding classifiers that don't win competitio */
    float prdTaxMx ;			/* Maximum Tax (Rate) on classifiers that produce a messages -- once/message. */
    float prdTaxPw ;			/* Power used to bias ProdTax--see BBLEARN.C */

    float totCCSup;				/* Total support for current candidate classifiers. */
    float aveCCSup;

    float totBBid;				/* Total 'biased' bids for current candidate classifiers. */
    float totCCBid;				/* Total of all bids by current candidates (NOT biased) */
    float hiCCBid;				/* Highest bid by current candidates (NOT biased)) */
    float lowCCBid;				/* Lowest bid by current candidates (NOT bissed) */

    /* Effector/detector interface objects: */
    unsigned int nmEffs ;		/* Number of effectors. */
    unsigned int detectRt ;		/* Detector rate - get messages from the environment */
						/*   every DetectRt steps of the major-cycle. */
    unsigned int effectRt ;		/* Effector rate - generate behavior from the message list */
    unsigned int detMsgMn ;		/* Minimum amount of space on MsgLists reserved for detector-messages. */

    unsigned int tOTEMtch ;		/* Total number msgs that matched effectors (before conflict resolution) */
    unsigned int tOTEAMsg ;		/* Total number msgs that activated effectors (after conflict resolution) */
    unsigned int tOTEAct ;		/* Total number times effectors activated */

    unsigned int madeMstk ;		/* set TRUE by domain-dependent GtSysRew() when mistake is made.
					This is used by GtSysRew() [domain dependent subroutine] and
					by DscCEff() in DSC-OPS.C . */

    /* Misc. other variables: */
    short dsplMsBl ;			/* Flag ON means message-strings displayed with blank every 4 loci */
    short dsplCfBl ;			/* Flag ON means classier-strings displayed with blank every 4 loci */

    short dsCfLst;			/* is '1' when selective display is on, else is '0' */

    /********** From here on we will have the discovery variables from dsclearn.def: ********/
    unsigned int dscDemo ;		/* Control 'discovery demonstration' disply. 0 is off. */
    unsigned int tOTNmOfs ;		/* Total number of offspring produced from this classifier */
    float mutdSpec ;			/* Amount specificity changes when one trit changes */
    /* variables for choosing 'parent' classifiers for the 'background Genetic Algorithm' operator */
    float frNewCfs ;			/* Fraction of classifier list to replace with new classifiers
					whenever 'background GA' operator is invoked.
					The "Generation Gap" of DeJong and others. */
    unsigned int mxNewCfs ;		/* Maximum number of new classifiers (and max to replace) at a step.
							This is just NmCfs * FrNewCfs. */
    unsigned int prPoolSz;
    unsigned int nxtPPool;
    short pkPrnWOR ;		/* 1 => pick parents with out replacement */
    unsigned int bidFit ;		/* 1 => Use BidRatio*Strength as fitness instead of Strength only */
    unsigned int nmNewCfs ;		/* Actual number of new classifiers during a step.
					This is index into NewCfs array for place to put next new classifier, too. */
    float tPrtBStr ;		/* Total 'Parent' Biased Strength of all classifiers, for chosing Parents. */
    float prntBPow ;		/* Power to raise strength to when computing 'biased' str for selecting parents */

    unsigned int fitnessCalc;		/* set true when ReCalcFitness() called, set False when Discover() called. */
    unsigned int cLSorted;		/* Set true when ReSrtCfs() called, to indicate list sorted this step. */
    /* Variables to control the strength of new classifiers */
    float nwCfStrF ;		/* New cf strength = NwCfStrF * AvrCfStr */
    unsigned int nwCfStrV ;	/* New cf strength is Fraction of Average. */

    /* variables for choosing classifiers to be replaced */
    unsigned int rCPoolSz ;		/* Number actually in the pool */
    unsigned int crowdFac ;		/* Crowding factor. This is number of classifiers in CrowdCfs[] (in order 0...)
					to examine and choose from when selecting a classifier to replace. */
    unsigned int randRplc ;		/* 1 => Pick classifiers to be replaced randomly (uniform over all cfs).
					0 => Pick classifier using RBiasStr/TRplBStr as probability. */
    float tRplBStr ;			/* Total "Replacable" Biased Strength for classifiers, for chosing Replacable cfs */
    float rplcBPow ;		/* Power to raise 1/strength to when computing 'biased' str for replacing cfs */
    unsigned int nmEligRp ;		/* number cfs eligible for replacement */
    float rplCfUBd ;		/* Don't replace cfs with S > AveS * RplCfUBd */
    float rplACUBd ;		/* Don't replace 'active' cfs with S > AveS * RplCfUBd */
    unsigned int rplACSBd ;		/* Classifier considered 'active' wrt RplACUBd
					RplACUBd if its StpLWon > CycleStp - RplACSBd */
    unsigned int mxCfCopy ;		/* if > 0, this is maximum number of copies of one classifier genotype */
    /*  Variables for the 'Background Genetic Algorithm' (reproduction/crossover) operator.
		BkgGARt	probability that operator will be invoked each step.
				This determines the background rate of reproduction.
		The following 3 variables define a cummulative prob. distr. which is used
		to detemine, for each pair of offspring, whether or not to cross them, and if so,
		what kind of cross to do.
		NOTE WELL:	The variables are tested in the order defined below.
					Also, it is currently the users responsibility to set these so they
					indeed define a cummulative prob. function.
		BGASCPr		Probability cross in one string (first or second condition, or action).
		BGAFCPr		Probability cross full classfier.
		BGANCPr		Probability no cross (just reproduce parents).
		Note: The following variables determine source of parents for BkgGA:
		BGABPPr		Probability parents from hi bidders.
					(Otherwise from Hi Str classifiers)    */
    float bkgGARt ;			/* Default is off */
    float bGASCPr ;
    float bGAFCPr ;
    float bGANCPr ;

    float bGABPPr ;

    unsigned int totBkgGA ;		/* Total num of times DscBkgOp invoked */
    unsigned int totBGASC ;		/* Total num of 'string' crosses */
    unsigned int totBGAFC ;		/* Total num of 'full' crosses */
    unsigned int totBGANC ;		/* Total num of bkg. reprod. with no cross */
    unsigned int totBGABP ;		/* Total num of times bidding parents used */

    /*  Variables for the 'Mutation' operator.
		The following probabilities define a cummulative prob. function
		over the types of mutation, used to decide what to mutate.
		Be sure these are in increasing order, and the last is 1.00
		**** THIS SHOULD BE FIXED UP: maybe use an array  *****	    */
    float muPrTot ;			/* Probability a mutation of some kind will occur. */

    /*	If mutation does occur in a classifier, the following
		values will be used to determine what will be mutated to what.*/
    float muPrCnd1 ;			/* Probability Mutate Condition 1 of a classifier */
    float muPrCnd2 ;			/* Probability Mutate Condition 2 of a classifier */
    float muPrAct ;			/* Probability Mutate Action string of a classifier */
    float muPrCT2 ;			/* Probability Mutate Condition 2 Type (MATCH-NOTMATCH) */
    float muPrAT ;			/* Probability Mutate classifier operator (ActType) of a classifier */

    /* The following 2 probabilities bias the mutation toward certain outcomes.
		For example, if a condition type is mutated, the result is NOT just
		obtained by changing MATCH to NOTMATCH (or vice versa). Instead, the
		new condition type is selected by a random draw under the the distribution
		defined by MuFrMC, i.e., that is the fraction that are MATCH, and the rest are NOTMATCH. */
    float muFrMC ;			/* Fraction of 'match' conditions by mutation */
    float muFrNSL ;			/* Fraction of non-specific loci (DON'T CARES) generated by mutation */
    unsigned int totMu ;			/* Total number of times mutation operator used in a run */

    /* Variables for the 'Cover Detector Messages' operator */
    unsigned int cDMsgs ;		/* Cover Detector Messages operator version: mine (1) or George's (2) */
    short cDMsgAct ;		/* set to 1 to create random actions by CDMsgs operator */
    float cDMsgsRt ;			/* Probability applied when on and triggered */
    unsigned int totCDM ;		/* Total number of times CDM operator used in a run (number of cfs generalized) */
    unsigned int totCDMC ;		/* Total number of loci generalized */
    unsigned int totCDML ;		/* Total number of conditions generalized */

    /*  Variables for the 'Cover Effectors' operator  */
    unsigned int cEffs ;		/* Number of classifiers to be generated at one application */
    float cEffsRt ;			/* If CEffs > 0, this is prob. operator applied when MadeMstk */
    short cEffSplz ;			/* Control "specialization" of new cover effector classifiers */
    unsigned int totCEf ;		/* Total number of times this operator used in a run */
    unsigned int totCEfW ;		/* Number of times parent from wrong predictors */
    unsigned int totCEfB ;		/* Number of times parents from bidders (not guessers) */

    /* Variables for 'Asunchronous Coupling Profitable Classifiers' operator */
    float aCPCRt ;			/* Prob applied on each step */
    unsigned int totACPC ;		/* Total number of times applied */
    float aCPCt ;			/* Threshold for "profit" (* cf strength) */
    float nwCCStrF ;			/* Determines strength of new coupled classifiers */

    unsigned int actCfPNm ;

    /* Variables for special creating and manipulating "Tag" regions */
    unsigned int tagLMx ;		/* Restrict creation of "tags" to loci from TagLMx to */
    unsigned int tagRMx ;		/* TagRMx, counting 0 as left locus. */

    /* Variables for the 'Couple Stage Setters' operator */
    float cSSRt ;			/* Prob apply at each step  */
    unsigned int totCSS ;		/* total number of times applied */
    float cSSFrM ;			/* Fraction of CSS couplings forced to use "memory" (11) messages */
    float cSSFrRf ;			/* Fraction of CSS couplings that create "reflex" classifiers */

    /* Variables for the 'Trigger on Low Bid' operator */
    float tLBRt ;			/* Prob apply at each step  */
    unsigned int totTLB ;		/* total number of times applied */
    float tLBFrE ;			/* Fraction of TLB couplings forced to use "effector" (10) messages */
    float tLBt ;			/* Threshold multiplier */
};    /* struct cfsc_s */

/* A Default cfsc_c structure: */
#define DEFAULT_cfsc_s { 1, 1, 1,\
    DEMOFF, \
    FALSE, \
    0,0, \
    FALSE,FALSE, \
    0,ADOFF,DEMOFMT1,ADOFF,DEMOFMT1, \
    ADOFF,DEMOFMT2,ADOFF,DEMOFMT1, \
    ADOFF,DEMOFMT1,ADOFF,DEMOFMT1, \
    ADOFF,1,		    /* SaveInt, SaveFNum */ \
    { NULL },			    /* DetMsgs (At least one initializer) */ \
    NULL,NULL, \
    0,0,0, \
    -1,			    /* LastCMsg */ \
    NULL,NULL, \
    0,0, \
    -1,			    /* LastNMsg */ \
    NULL,NULL, \
    0,0, \
    -1,			    /* LastOMsg */ \
    30,2,50, \
    0,0,0,0, \
    NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL, \
    0,0,0, \
    1,0, \
    CFLSTSZ,		    /* NmCfsMx */ \
    0,0,0, \
    0,0,0,0,		    /* TotCfStr, ... ,AveCfBR */ \
    1000,100,99999,0.1,0.1, \
    0.95,0.20,		    /* GRCFrMC, GRCFrNSL */ \
    0.0, \
    0,0,0,0,0,0,0,0,0, \
    0,			    /* NmMtchs */ \
    0,0,0,0,		    /* NmIEMMx, ... , NmIIMsgs */ \
    0, 200000.0, 0, 1.00, 1.0, 1.0, 0.0, 0, 0, 1,   /* last is SuppBid */ \
    1.0, 0.10, 1.0, 0, 0, 0, 0.0005, 0.00, 0.005, 0.00, 0.000, 1.50, /* last is PrdTaxPw */ \
    0.0, 0.0,		    /* TotCCSup, AveCCSup */ \
    0.0, 0.0, 0.0, 0.0, \
    0, 1, 1, 4, \
    0, 0, 0,		    /* last is TOTEAct */ \
    0,			    /* MadeMstk */ \
    ON, ON,		    /* last is DsplCfBl */ \
    0,	\
    /********* From here on starts the dsclearn section: *****************/ \
    0, 0, \
    0, /* MutdSpec */  \
    0.04, 0, \
    0, 0, 0, 0, 0,	    /* last is NmNewCfs */ \
    0.0, 1.0, \
    0, 0, \
    1.0, NWSTRFA, 	    /* last is NwCfStrV */ \
    0, 1, 0, \
    0, 1.0, 0, 1.0, 1000.0, 0, \
    0,			    /* MxCfCopy */ \
    0.0, 0.0, 0.5, 1.0, 0.0, \
    0, 0, 0, 0, 0,	    /* last is TotBGABP */ \
    0.04, \
    0.32, 0.64, 0.96, 1.00, 1.00, \
    0.95, 0.20, 0,	    /* last is TotMu */ \
    0, 0, 0.0, 0, 0, 0,	\
    0, 0.0, 0, 0, 0, 0,	    /* last is TotCEfB */ \
    0.0, 0, 0.0, 0.0, \
    0,			    /* ActCfPNm */ \
    2,11, \
    0.0, 0, 0.0, 0.0,	    /* last is CSSFrRf */ \
    0.0, 0, 0.0, 0.0	    /* last is TLBt */ \
}
    

/* Defines to enable using the entries like global variables: */
#define CFSC_VAR    Cfsc
#define SpecPay	    (CFSC_VAR->specPay)
#define URndSd	    (CFSC_VAR->uRndSd)
#define CycleStp    (CFSC_VAR->cycleStp)
#define DemoLev	    (CFSC_VAR->demoLev)
#define EchoFlg	    (CFSC_VAR->echoFlg)
#define	LogFlg	    (CFSC_VAR->logFlg)
#define	GenFlg	    (CFSC_VAR->genFlg)
#define	ClampFlg    (CFSC_VAR->clampFlg)
#define	StopFlag    (CFSC_VAR->stopFlag)

#define	AutoDspl    (CFSC_VAR->autoDspl)
#define	ADCMInt     (CFSC_VAR->aDCMInt)
#define	ADCMFmt     (CFSC_VAR->aDCMFmt)
#define	ADDMInt     (CFSC_VAR->aDDMInt)
#define	ADDMFmt     (CFSC_VAR->aDDMFmt)
#define	ADCfInt     (CFSC_VAR->aDCfInt)
#define	ADCfFmt     (CFSC_VAR->aDCfFmt)
#define	ADVarInt    (CFSC_VAR->aDVarInt)
#define	ADVarFmt    (CFSC_VAR->aDVarFmt)
#define	ADCycInt    (CFSC_VAR->aDCycInt)
#define	ADCycFmt    (CFSC_VAR->aDCycFmt)
#define	ADEnvInt    (CFSC_VAR->aDEnvInt)
#define	ADEnvFmt    (CFSC_VAR->aDEnvFmt)

#define	SaveInt     (CFSC_VAR->saveInt)
#define	SaveFNum    (CFSC_VAR->saveFNum)
#define	DetMsgs     (CFSC_VAR->detMsgs)

#define	CurMsgs     (CFSC_VAR->curMsgs)
#define	NextCMsg    (CFSC_VAR->nextCMsg)
#define	NmCMsgs     (CFSC_VAR->nmCMsgs)
#define	NmCMsgIn    (CFSC_VAR->nmCMsgIn)
#define	NmDetMsg    (CFSC_VAR->nmDetMsg)
#define	LastCMsg    (CFSC_VAR->lastCMsg)

#define	NewMsgs     (CFSC_VAR->newMsgs)
#define	NextNMsg    (CFSC_VAR->nextNMsg)
#define	NmNMsgs     (CFSC_VAR->nmNMsgs)
#define	NmNMsgIn    (CFSC_VAR->nmNMsgIn)
#define	LastNMsg    (CFSC_VAR->lastNMsg)

#define	OldMsgs    (CFSC_VAR->oldMsgs)
#define	NextOMsg    (CFSC_VAR->nextOMsg)
#define	NmOMsgs    (CFSC_VAR->nmOMsgs)
#define	NmOMsgIn    (CFSC_VAR->nmOMsgIn)
#define	LastOMsg    (CFSC_VAR->lastOMsg)

#define	NmIMsgMx    (CFSC_VAR->nmIMsgMx)
#define	NmDMsgMx    (CFSC_VAR->nmDMsgMx)
#define	MsgIntDf    (CFSC_VAR->msgIntDf)

#define	TOTMtch     (CFSC_VAR->tOTMtch)
#define	TOTMsPrd    (CFSC_VAR->tOTMsPrd)
#define	TOTMsPst    (CFSC_VAR->tOTMsPst)
#define	TOTDtMsg    (CFSC_VAR->tOTDtMsg)

#define	CfLst       (CFSC_VAR->cfLst)
#define	CurCfs      (CFSC_VAR->curCfs)
#define	NxtFrCf     (CFSC_VAR->nxtFrCf)
#define	CandCfs     (CFSC_VAR->candCfs)
#define	LstCndCf    (CFSC_VAR->lstCndCf)
#define	StrSrtCf    (CFSC_VAR->strSrtCf)
#define	HiStrCf     (CFSC_VAR->hiStrCf)
#define	HighestStrCf    (CFSC_VAR->highestStrCf)
#define	LowestStrCf     (CFSC_VAR->lowestStrCf)

#define	OneMPerC    (CFSC_VAR->oneMPerC)
#define	DelHall    (CFSC_VAR->delHall)
#define	MxDupMsg    (CFSC_VAR->mxDupMsg)

#define	NxtCfId    (CFSC_VAR->nxtCfId)
#define	NmCfs    (CFSC_VAR->nmCfs)
#define	NmCfsMx    (CFSC_VAR->nmCfsMx)
#define	NmCandCf    (CFSC_VAR->nmCandCf)
#define	NmCfWon    (CFSC_VAR->nmCfWon)
#define	NmCfPost    (CFSC_VAR->nmCfPost)

#define	TotCfStr    (CFSC_VAR->totCfStr)
#define	AveCfStr    (CFSC_VAR->aveCfStr)
#define	TotCfBR    (CFSC_VAR->totCfBR)
#define	AveCfBR    (CFSC_VAR->aveCfBR)

#define	CfStrDf    (CFSC_VAR->cfStrDf)
#define	CfSuppDf    (CFSC_VAR->cfSuppDf)
#define	CfStrMax    (CFSC_VAR->cfStrMax)
#define	CfStrMin    (CFSC_VAR->cfStrMin)
#define	CfBidMin    (CFSC_VAR->cfBidMin)

#define	GRCFrMC    (CFSC_VAR->gRCFrMC)
#define	GRCFrNSL    (CFSC_VAR->gRCFrNSL)

#define	TOTBidRa    (CFSC_VAR->tOTBidRa)
#define	TOTNmBid    (CFSC_VAR->tOTNmBid)
#define	TOTNmWin    (CFSC_VAR->tOTNmWin)
#define	TOTCfPst    (CFSC_VAR->tOTCfPst)
#define	TOTBGStr    (CFSC_VAR->tOTBGStr)
#define	TOTActEf    (CFSC_VAR->tOTActEf)
#define	TOTNmPRw    (CFSC_VAR->tOTNmPRw)
#define	TOTNmNRw    (CFSC_VAR->tOTNmNRw)
#define	TOTSyRwP    (CFSC_VAR->tOTSyRwP)
#define	TOTSyRwN    (CFSC_VAR->tOTSyRwN)

#define	NmMtchs    (CFSC_VAR->nmMtchs)

#define	NmIEMMx    (CFSC_VAR->nmIEMMx)
#define	NmIEMsgs    (CFSC_VAR->nmIEMsgs)
#define	NmIIMMx    (CFSC_VAR->nmIIMMx)
#define	NmIIMsgs    (CFSC_VAR->nmIIMsgs)

#define	BBType    (CFSC_VAR->bBType)
#define	SysTreas    (CFSC_VAR->sysTreas)
#define	SysRew    (CFSC_VAR->sysRew)
#define	FrPayDet    (CFSC_VAR->frPayDet)
#define	FrPayNEC    (CFSC_VAR->frPayNEC)
#define	DMShare    (CFSC_VAR->dMShare)
#define	DMDscnt    (CFSC_VAR->dMDscnt)
#define	ShareRew    (CFSC_VAR->shareRew)
#define	AllPayBd    (CFSC_VAR->allPayBd)
#define	SuppBid    (CFSC_VAR->suppBid)

#define	BidPow    (CFSC_VAR->bidPow)
#define	Bid_k    (CFSC_VAR->bid_k)
#define	BRPow    (CFSC_VAR->bRPow)
#define	EBRPow    (CFSC_VAR->eBRPow)
#define	BtCSB    (CFSC_VAR->btCSB)
#define	BtCMB    (CFSC_VAR->btCMB)
#define	HeadTax    (CFSC_VAR->headTax)
#define	AHeadTax    (CFSC_VAR->aHeadTax)
#define	BidTax    (CFSC_VAR->bidTax)
#define	FBidTax    (CFSC_VAR->fBidTax)
#define	PrdTaxMx    (CFSC_VAR->prdTaxMx)
#define	PrdTaxPw    (CFSC_VAR->prdTaxPw)

#define	TotCCSup    (CFSC_VAR->totCCSup)
#define	AveCCSup    (CFSC_VAR->aveCCSup)
#define	TotBBid    (CFSC_VAR->totBBid)
#define	TotCCBid    (CFSC_VAR->totCCBid)
#define	HiCCBid    (CFSC_VAR->hiCCBid)
#define	LowCCBid    (CFSC_VAR->lowCCBid)

#define	NmEffs    (CFSC_VAR->nmEffs)
#define	DetectRt    (CFSC_VAR->detectRt)
#define	EffectRt    (CFSC_VAR->effectRt)
#define	DetMsgMn    (CFSC_VAR->detMsgMn)
#define	TOTEMtch    (CFSC_VAR->tOTEMtch)
#define	TOTEAMsg    (CFSC_VAR->tOTEAMsg)
#define	TOTEAct    (CFSC_VAR->tOTEAct)
#define	MadeMstk    (CFSC_VAR->madeMstk)

#define	DsplMsBl    (CFSC_VAR->dsplMsBl)
#define	DsplCfBl    (CFSC_VAR->dsplCfBl)
#define	DsCfLst    (CFSC_VAR->dsCfLst)

/***** dsclearn section: *****/
#define DscDemo	    (CFSC_VAR->dscDemo)
#define TOTNmOfs    (CFSC_VAR->tOTNmOfs)
#define MutdSpec    (CFSC_VAR->mutdSpec)
#define FrNewCfs    (CFSC_VAR->frNewCfs)
#define MxNewCfs    (CFSC_VAR->mxNewCfs)
#define PrPoolSz    (CFSC_VAR->prPoolSz)
#define NxtPPool    (CFSC_VAR->nxtPPool)
#define PkPrnWOR    (CFSC_VAR->pkPrnWOR)
#define BidFit	    (CFSC_VAR->bidFit)
#define NmNewCfs    (CFSC_VAR->nmNewCfs)
#define TPrtBStr    (CFSC_VAR->tPrtBStr)
#define PrntBPow    (CFSC_VAR->prntBPow)
#define FitnessCalc (CFSC_VAR->fitnessCalc)
#define CLSorted    (CFSC_VAR->cLSorted)
#define NwCfStrF    (CFSC_VAR->nwCfStrF)
#define NwCfStrV    (CFSC_VAR->nwCfStrV)
#define RCPoolSz    (CFSC_VAR->rCPoolSz)
#define CrowdFac    (CFSC_VAR->crowdFac)
#define RandRplc    (CFSC_VAR->randRplc)
#define TRplBStr    (CFSC_VAR->tRplBStr)
#define RplcBPow    (CFSC_VAR->rplcBPow)
#define NmEligRp    (CFSC_VAR->nmEligRp)
#define RplCfUBd    (CFSC_VAR->rplCfUBd)
#define RplACUBd    (CFSC_VAR->rplACUBd)
#define RplACSBd    (CFSC_VAR->rplACSBd)
#define MxCfCopy    (CFSC_VAR->mxCfCopy)

#define BkgGARt	    (CFSC_VAR->bkgGARt)
#define BGASCPr	    (CFSC_VAR->bGASCPr)
#define BGAFCPr	    (CFSC_VAR->bGAFCPr)
#define BGANCPr	    (CFSC_VAR->bGANCPr)
#define BGABPPr	    (CFSC_VAR->bGABPPr)
#define TotBkgGA    (CFSC_VAR->totBkgGA)
#define TotBGASC    (CFSC_VAR->totBGASC)
#define TotBGAFC    (CFSC_VAR->totBGAFC)
#define TotBGANC    (CFSC_VAR->totBGANC)
#define TotBGABP    (CFSC_VAR->totBGABP)

#define MuPrTot	    (CFSC_VAR->muPrTot)
#define MuPrCnd1    (CFSC_VAR->muPrCnd1)
#define MuPrCnd2    (CFSC_VAR->muPrCnd2)
#define MuPrAct	    (CFSC_VAR->muPrAct)
#define MuPrCT2	    (CFSC_VAR->muPrCT2)
#define MuPrAT	    (CFSC_VAR->muPrAT)
#define MuFrMC	    (CFSC_VAR->muFrNSL)
#define MuFrNSL	    (CFSC_VAR->muFrNSL)
#define TotMu	    (CFSC_VAR->totMu)

#define CDMsgs	    (CFSC_VAR->cDMsgs)
#define CDMsgAct    (CFSC_VAR->cDMsgAct)
#define CDMsgsRt    (CFSC_VAR->cDMsgsRt)
#define TotCDM	    (CFSC_VAR->totCDM)
#define TotCDMC	    (CFSC_VAR->totCDMC)
#define TotCDML	    (CFSC_VAR->totCDML)

#define CEffs	    (CFSC_VAR->cEffs)
#define CEffsRt	    (CFSC_VAR->cEffsRt)
#define CEffSplz    (CFSC_VAR->cEffSplz)
#define TotCEf	    (CFSC_VAR->totCEf)
#define TotCEfW	    (CFSC_VAR->totCEfW)
#define TotCEfB	    (CFSC_VAR->totCEfB)

#define ACPCRt	    (CFSC_VAR->aCPCRt)
#define TotACPC	    (CFSC_VAR->totACPC)
#define ACPCt	    (CFSC_VAR->aCPCt)
#define NwCCStrF    (CFSC_VAR->nwCCStrF)
#define ActCfPNm    (CFSC_VAR->actCfPNm)

#define TagLMx	    (CFSC_VAR->tagLMx)
#define TagRMx	    (CFSC_VAR->tagRMx)

#define CSSRt	    (CFSC_VAR->cSSRt)
#define TotCSS	    (CFSC_VAR->totCSS)
#define CSSFrM	    (CFSC_VAR->cSSFrM)
#define CSSFrRf	    (CFSC_VAR->cSSFrRf)

#define TLBRt	    (CFSC_VAR->tLBRt)
#define TotTLB	    (CFSC_VAR->totTLB)
#define TLBFrE	    (CFSC_VAR->tLBFrE)
#define TLBt	    (CFSC_VAR->tLBt)

/* ------------------------------- */
#endif  /* _CFSCTYPES_H */
